#!/usr/bin/python

PROGRAM_NAME = 'lazymake.py'
PROGRAM_VERSION = '0.1.1'

LAZY65 = 'lazy65'
CA65 = 'ca65'
LD65 = 'ld65'

import multiprocessing
MAX_PROCESSES = multiprocessing.cpu_count() #speed up compilation on multi-core systems

import os, sys
import argparse
import shutil
import subprocess, shlex, time

WINDOWS = (os.name=='nt')

class ProcessManager:
	def __init__(self, maxProcesses):
		self.processes = []
		self.maxProcesses = maxProcesses
		pass
		
	def add(self, command):
		#args = shlex.split('python '+command)
		p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		self.processes.append((p,command))
		#we will print command, stdout and stderr when the process has finished executing
		return self.wait(self.maxProcesses)
		
	def wait(self, nProcesses):
		#block until a new process can be run
		while len(self.processes) >= nProcesses:
			if WINDOWS:
				time.sleep(.01)
			else:
				os.wait()
			for pt in self.processes[:]:
				p = pt[0]
				res = p.poll()
				if res!=None:
					command = pt[1]
					print command
					out, err = p.communicate()
					errcode = p.returncode
					if out:
						print out
					sys.stderr.write(err)
					self.processes.remove(pt)
					if res!=0:
						return 1
		return 0
		
	def waitAll(self):
		return self.wait(1)
		
		
processManager = ProcessManager(MAX_PROCESSES)


def clean(args):
	try:
		if os.path.isfile(args.outfile):
			print 'deleting '+args.outfile
			os.remove(args.outfile)
	except Exception, e:
		print e
		return 1
	folder = args.objdir
	for file in os.listdir(folder):
		path = os.path.join(folder, file)
		try:
			if os.path.isfile(path):
				print 'deleting '+path
				os.unlink(path)
			elif os.path.isdir(path):
				print 'deleting '+path
				shutil.rmtree(path)
		except Exception, e:
			print e
			return 1
	return 0
	
def all(args):
	if os.path.exists(args.outfile):
		outfileMTime = os.path.getmtime(args.outfile)
	else:
		outfileMTime = None
		
	if not os.path.isdir(args.srcdir):
		print 'source directory not found: "'+args.srcdir+'"'
		return
	
	buildTime = time.time()
	
	dependencies = {}
	modules = {}
	allModules = {}
	importedModules = {}
	#add changed modules to compile list
	for path, subdirs, files in os.walk(args.srcdir):
		for filename in files:
			f = os.path.join(path, filename)
			moduleName = f[len(args.srcdir)+1:]
			dot = moduleName.rfind('.')
			ext = ''
			if dot!=-1:
				ext = moduleName[dot:]
				moduleName = moduleName[:dot]
			if ext=='.z':
				#./Module.z -> Module
				#./Module/__root__.z -> Module
				if moduleName[-8:] == '__root__':
					moduleName = moduleName[:-9]
				moduleName = moduleName.replace('\\','.')
				moduleName = moduleName.replace('/','.')
				if outfileMTime==None or os.path.getmtime(f)>outfileMTime:
					modules[moduleName] = True
				allModules[moduleName] = True
				dependencies[moduleName] = readDependencies(args.objdir, moduleName)
				
	
	#add modules to compile list that are dependent on changed modules
	buildAll = False
	for moduleName in allModules:
		deps = dependencies[moduleName]
		if deps==None:
			#if even a single dependency file is missing we need to rebuild all sources
			buildAll = True
			break
		else:
			for dm in deps:
				if dm in modules:
					modules[moduleName] = True
					#print 'Added '+moduleName+' to build list because it depends on the changed '+dm
				
	if buildAll:
		modules = allModules
		importedModules = allModules
		
	for moduleName in modules:
		if dependencies[moduleName]==None:
			continue
		for dm in dependencies[moduleName]:
			importedModules[dm] = True
		
	if len(modules)==0:
		print PROGRAM_NAME+": nothing to be done for 'all'."
		return
	
	#step 1 - lazy65 1st pass
	#build symbol files in advance for used modules
	for moduleName in importedModules:
		command = LAZY65+' --sym-only -s "'+args.srcdir+'" -o "'+args.objdir+'" '+moduleName
		res = processManager.add(command)
		if res!=0:
			printError(res)
			return res
	res = processManager.waitAll()
	if res!=0:
		printError(res)
		return res
	
	#step 2 - lazy65 2nd pass
	for moduleName in modules:
		#TODO: for modules handled in step 1, run 2nd pass only
		command = LAZY65+' -d -t '+str(buildTime)+' -s "'+args.srcdir+'" -o "'+args.objdir+'" '+moduleName
		res = processManager.add(command)
		if res!=0:
			printError(res)
			return res
	res = processManager.waitAll()
	if res!=0:
		printError(res)
		return res
	
	#step 3 - ca65
	for moduleName in modules:
		modulePath = moduleName.replace('.','/')
		source = args.objdir+'/'+modulePath+'.s'
		object = args.objdir+'/'+modulePath+'.o'
		command = CA65+' "'+source+'" -o "'+object+'"'
		res = processManager.add(command)
		if res!=0:
			printError(res)
			return res
	res = processManager.waitAll() #TODO: multiple ca65 processes OK? (check SBP source) Have option to disable multi-processes for ca65 jobs?
	if res!=0:
		printError(res)
		return res
	
	#step 4 - ld65
	objects = ''
	for moduleName in allModules:
		modulePath = moduleName.replace('.','/')
		objects += '"'+args.objdir+'/'+modulePath+'.o" '
	command = LD65+' -o "'+args.outfile+'" -C '+args.config+' ' + objects
	print command
	if os.system(command)!=0:
		printError(2)
		return 2
	
	print args.outfile+': Build complete'
	
	return 0
	
def readDependencies(objdir, moduleName):
	file = objdir + '/' + moduleName.replace('.','/') + '.d'
	deps = None
	if os.path.isfile(file):
		with open(file, 'r') as f:
			lines = f.readlines()
			if len(lines)>=1:
				line = lines[0].rstrip()
				deps = line.split(' ')
				if deps==['']:
					deps = []
			f.close()
	return deps
	
def printError(res):
	print PROGRAM_NAME+': Error '+str(res)

def main():
	ap = argparse.ArgumentParser(prog=PROGRAM_NAME, usage='%(prog)s [options] [target] [-o output file]')
	ap.add_argument("target", help="clean / all", default='all', nargs='?')
	ap.add_argument("-s", "--srcdir", help="Specify source directory (default: 'src')", default='src')
	ap.add_argument("-d", "--objdir", help="Specify object file directory (default: 'obj')", default='obj')
	ap.add_argument("-o", "--outfile", help="Specify output file", default=os.path.basename(os.getcwd())+'.nes')
	ap.add_argument("-v","--verbose", help="", action="store_true")
	ap.add_argument("-j","--jobs", help="Number of commands to run simultaneously", default=0, type=int)
	ap.add_argument("-C", "--config", help="Linker config file for ld65 (default: 'linker.cfg')", default='linker.cfg')
	ap.add_argument("--version", help="Print version number", action="store_true")
	
	args = ap.parse_args()
	
	if args.version:
		print PROGRAM_NAME + ' V' + PROGRAM_VERSION
		return
	
	if args.srcdir == None:
		args.srcdir = './'
	if args.objdir == None:
		args.objdir = 'obj'
	
	if args.jobs!=0:
		processManager.maxProcesses = args.jobs
	
	if args.target=='clean':
		return clean(args)
	elif args.target=='all':
		return all(args)
	#add additional targets here
	else:
		print 'invalid target "'+args.target+'"'
		return 1


sys.exit( main() )
